#####################################################################
# CLASS Agent_pair
#####################################################################
#
# Contoller of Agent pairs arriving wrt random process.
#
# VARIABLE:
#
# snode dnode: source node, destination node
# ctype: connection type (TCP_pair)
#
# pid: pair id, id of current Agent_pair, all connections within has the same $pid (src->dst)
#
# conns: array of connections of type $ctype
# conns_per_pair: number of connections in this pair
#
# logfile: log file
#
# rv_intval: (r.v.) flow arrival interval
# rv_nbytes: (r.v.) flow size in bytes
#
# tnext: (memory) next flow start time
# actfl: (memory) number of active flows in this pair
#
#
# FUNCTION:
#
# --- initializing ---
# init { args }
# setup { srcnode dstnode pairid connpp conntype }
#
# --- helper ---
# attach-logfile { logf }
#
# --- schedule ---
# pass global $isStable to TCP_pair
# init_schedule {} ==> schedule{} 
# schedule { connid } ==> check_if_behind{}
# check_if_behind {} ==> schedule{}
#
# --- random process RV initialization ---
# set_PParrival_process { lambda mean_nbytes shape rands1 rands2 }
# set_PEarrival_process { lambda mean_nbytes rands1 rands2 } 
# set_PCarrival_process { lambda cdffile rands1 rands2 }
# set_LCarrival_process { lambda cdffile rands1 rands2 }
# set_PBarrival_process { lambda mean_nbytes nbytes1 nbytes2 rands1 rands2 }
# set_PFarrival_process { lambda mean_nbytes rand1 }
#
# --- callback ---
# fin_notify { connid dt taskid stable } ==> schedule{}
# start_notify { connid npkts taskid stable }
#
#####################################################################

Class Agent_pair

#:::::::::::::::: initializing ::::::::::::::::

Agent_pair instproc init { args } {
    eval $self next $args

    $self instvar tnext actfl
    $self set tnext 0.0
    $self set actfl 0

    # p2-task (GONG)
    # $self instvar available
}

Agent_pair instproc setup { srcnode dstnode pairid connpp conntype } { 
    # PUBLIC
    # initialize pair id $pid, create and initialize $connpp $conntype (TCP_pair) connections between $srcnode and $dstnode

    global fid_
    $self instvar snode dnode ctype
    $self instvar pid
    $self instvar conns conns_per_pair
    
    $self set snode $srcnode
    $self set dnode $dstnode
    $self set ctype $conntype
    $self set pid $pairid 
    $self set conns_per_pair $connpp
    
    for { set i 0 } { $i < $conns_per_pair } { incr i } {
        $self set conns($i) [new $conntype]
        $conns($i) setup $snode $dnode 
        $conns($i) setpid $pid
        $conns($i) setconnid $i
        $conns($i) setfid $fid_

        # p2-task (GONG)
        # $self set available($i) 1

        incr fid_
    }
}

#:::::::::::::::: helper ::::::::::::::::

Agent_pair instproc attach-logfile { logf } {
    $self instvar logfile
    $self set logfile $logf
}

#:::::::::::::::: schedule ::::::::::::::::

Agent_pair instproc init_schedule {} {
    # PUBLIC
    # set callback and schedule all connections according to the arrival process (must be called after set_**arrival_process{})

    global ns
    $self instvar conns conns_per_pair 
    $self instvar rv_intval
    $self instvar tnext
    
    # p1-flow-based task (GONG)
    # $self instvar nfpt rv_nfpt
    # $self set nfpt [expr int(ceil([$rv_nfpt value]))]

    $self set tnext [expr [$ns now] + [$rv_intval value]]

    # p1-flow-based task (GONG)
    # $self instvar task_id
    # $self set task_id [expr int($tnext * 1000)] 
    
    for { set i 0 } { $i < $conns_per_pair } { incr i } {
        $conns($i) set_callback $self start_notify fin_notify
        $self schedule $i
    }
}

# p2-task (GONG)
# Agent_pair instproc schedule_ext { taskctrl genfunc finfunc time nbytes tagname taskprio } {

#     global ns
#     $self instvar available conns

#     for { set i 0 } { $i < [array size conns] } { incr i } {
#         if { $available($i) > 0 } {
#             # $self start_notify is dummy here
#             $conns($i) set_callback_ext $self start_notify fin_notify_ext $taskctrl $genfunc $finfunc
#             $ns at $time "$conns($i) start_ext $nbytes $tagname $taskprio" 
#             $self set available($i) 0
#             break
#         }
#     }
    
#     if { $i == [array size conns] } {

#         global fid_
#         $self instvar snode dnode ctype
#         $self instvar pid
        
#         $self set conns($i) [new $ctype]
#         $conns($i) setup $snode $dnode
#         $conns($i) setpid $pid
#         $conns($i) setconnid $i
#         $conns($i) setfid $fid_
#         incr fid_
        
#         $conns($i) set_callback_ext $self start_notify fin_notify_ext $taskctrl $genfunc $finfunc
#         $ns at $time "$conns($i) start_ext $nbytes $tagname $taskprio" 
#         $self set available($i) 0
#     }
# }

Agent_pair instproc schedule { connid } {
    # PRIVATE
    # schedule connection $connid at $tnext time (given by init_schedule{} or schedule{} from previous call or other connections in the same pair), and generate next arrival time 

    global ns var
    $self instvar conns
    $self instvar rv_intval rv_nbytes
    $self instvar tnext

    # ??? don't know why this would occur
    if { [$ns now] > $tnext } {
        return
        puts "Error, Not enough flows ! Aborting!"
        flush stdout
        exit 
    }

    # p1-flow-based task (GONG)
    # $self instvar nfpt task_id rv_nfpt
    # if { $nfpt == 0 } {
    #     $self set nfpt [expr int(ceil([$rv_nfpt value]))]
    #     $self set task_id [expr int($tnext * 1000)] 
    # } else {
    #     incr nfpt -1 
    # }
    # CAUTION: cdf is in pkts, so rv_nbytes is in pkts
    
    # p2-task (GONG)
    # $ns at $tnext "$conns($connid) start [expr ceil ([$rv_nbytes value])] $task_id"
    
    # load stable (GONG)
    global isStable 
    
    # load stable (GONG)
    # $ns at $tnext "$conns($connid) start [expr ceil ([$rv_nbytes value])] $task_id $isStable"
    # task_id = 0 to not running flow-based task
    $ns at $tnext "$conns($connid) start [expr ceil ([$rv_nbytes value])] 0 $isStable" 

    $self set tnext [expr $tnext + [$rv_intval value]]
    $ns at [expr $tnext - 0.0000001] "$self check_if_behind"
}

Agent_pair instproc check_if_behind {} {
    
    global ns var flow_gen
    $self instvar conns
    $self instvar tnext

    # p2-task (GONG)
    # if { $tnext < [expr [$ns now] + 0.0000002] } {}
    if { $flow_gen < $var(sim_end) && $tnext < [expr [$ns now] + 0.0000002] } { 

        global fid_
        $self instvar snode dnode ctype
        $self instvar pid
        
        set i [array size conns]
        $self set conns($i) [new $ctype]
        $conns($i) setup $snode $dnode
        $conns($i) setpid $pid
        $conns($i) setconnid $i
        $conns($i) setfid $fid_
        incr fid_
        
        $conns($i) set_callback $self start_notify fin_notify

        $self schedule $i
    }
}

#:::::::::::::::: random process RV initialization ::::::::::::::::
# setup random variable rv_intval and rv_nbytes.
# get random value generated: $rv_intval value

Agent_pair instproc set_PParrival_process { lambda mean_nbytes shape rands1 rands2 } {
    # PUBLIC
    # Poisson/Pareto
    # arrival: poissson with rate $lambda
    # size: pareto with mean $mean_nbytes bytes and shape parameter $shape

    $self instvar rv_intval rv_nbytes

    set rng1 [new RNG]
    $rng1 seed $rands1
    $self set rv_intval [new RandomVariable/Exponential]
    $rv_intval use-rng $rng1
    $rv_intval set avg_ [expr 1.0 / $lambda]

    set rng2 [new RNG]
    $rng2 seed $rands2
    $self set rv_nbytes [new RandomVariable/Pareto]
    $rv_nbytes use-rng $rng2
    $rv_nbytes set avg_ $mean_nbytes; # pkt oriented hack [expr $mean_nbytes/1500]
    $rv_nbytes set shape_ $shape
}

Agent_pair instproc set_PEarrival_process { lambda mean_nbytes rands1 rands2 } {
    # PUBLIC
    # Poisson/Exponential
    # arrival: poissson with rate $lambda
    # size: exponential with mean $mean_nbytes bytes

    $self instvar rv_intval rv_nbytes

    set rng1 [new RNG]
    $rng1 seed $rands1
    $self set rv_intval [new RandomVariable/Exponential]
    $rv_intval use-rng $rng1
    $rv_intval set avg_ [expr 1.0 / $lambda]

    set rng2 [new RNG]
    $rng2 seed $rands2
    $self set rv_nbytes [new RandomVariable/Exponential]
    $rv_nbytes use-rng $rng2
    $rv_nbytes set avg_ $mean_nbytes
}

Agent_pair instproc set_PCarrival_process { lambda cdffile rands1 rands2 } {
    # PUBLIC
    # Poisson/CDF:
    # arrival: poisson with rate $lambda
    # size: custom defined expirical cdf from $cdffile

    $self instvar rv_intval rv_nbytes

    set rng1 [new RNG]
    $rng1 seed $rands1
    $self set rv_intval [new RandomVariable/Exponential]
    $rv_intval use-rng $rng1
    $rv_intval set avg_ [expr 1.0/$lambda]

    set rng2 [new RNG]
    $rng2 seed $rands2
    $self set rv_nbytes [new RandomVariable/Empirical]
    $rv_nbytes use-rng $rng2
    $rv_nbytes set interpolation_ 2
    $rv_nbytes loadCDF $cdffile
}

Agent_pair instproc set_LCarrival_process { lambda cdffile rands1 rands2 } {
    # PUBLIC
    # LogNormal/CDF
    # arrival: lognormal with rate $lambda
    # size: custom defined expirical cdf from $cdffile

    $self instvar rv_intval rv_nbytes

    set rng1 [new RNG]
    $rng1 seed $rands1
    $self set rv_intval [new RandomVariable/LogNormal]
    $rv_intval use-rng $rng1
    $rv_intval set avg_ [expr 0.5 * log(1.0 / ($lambda * $lambda) / 5.0)]
    $rv_intval set std_ [expr sqrt(log(5.0))]

    set rng2 [new RNG]
    $rng2 seed $rands2
    $self set rv_nbytes [new RandomVariable/Empirical]
    $rv_nbytes use-rng $rng2
    $rv_nbytes set interpolation_ 2
    $rv_nbytes loadCDF $cdffile
}

Agent_pair instproc set_PBarrival_process { lambda mean_nbytes nbytes1 nbytes2 rands1 rands2 } {
    # PUBLIC
    # Poisson/Binomial
    # arrival: poissson with rate $lambda
    # size: binomial with mean $mean_nbytes bytes flow size $nbytes1 and $nbytes2

    $self instvar rv_intval rv_nbytes

    set rng1 [new RNG]
    $rng1 seed $rands1
    $self set rv_intval [new RandomVariable/Exponential]
    $rv_intval use-rng $rng1
    $rv_intval set avg_ [expr 1.0 / $lambda]

    set rng2 [new RNG]
    $rng2 seed $rands2
    $self set rv_nbytes [new RandomVariable/Binomial]
    $rv_nbytes use-rng $rng2
    set p_ [expr (1.0 * $mean_nbytes - $nbytes2) / ($nbytes1 - $nbytes2)]
    if { $p_ < 0 } {
        puts "In PBarrival, prob for bimodal p_ is negative %p_, exiting.. "
        flush stdout
        exit 0
    }
    $rv_nbytes set p1_ $p_
    $rv_nbytes set s1_ $nbytes1
    $rv_nbytes set s2_ $nbytes2
}

Agent_pair instproc set_PFarrival_process { lambda mean_nbytes rand1 } {
    # PUBLIC
    # Poisson/Uniform
    # arrival: poissson with rate $lambda
    # size: unifom with $mean_nbytes bytes 

    $self instvar rv_intval rv_nbytes

    set rng1 [new RNG]
    $rng1 seed $rand1
    $self set rv_intval [new RandomVariable/Exponential]
    $rv_intval use-rng $rng1
    $rv_intval set avg_ [expr 1.0 / $lambda]

    set rng2 [new RNG]
    $rng2 seed 12345
    $self set rv_nbytes [new RandomVariable/Uniform]
    $rv_nbytes use-rng $rng2
    $rv_nbytes set min_ $mean_nbytes
    $rv_nbytes set max_ $mean_nbytes
}

# p1-flow-based task (GONG)
# Agent_pair instproc set_FlowPerTask_process { mean_nfpt shape rand } {
    
#     $self instvar rv_nfpt

#     set rng [new RNG]
#     $rng seed $rand
#     $self set rv_nfpt [new RandomVariable/Pareto]
#     $rv_nfpt use-rng $rng
#     $rv_nfpt set avg_ $mean_nfpt
#     $rv_nfpt set shape_ $shape
# }

#:::::::::::::::: callback ::::::::::::::::

Agent_pair instproc start_notify { connid npkts taskid stable } {
    # CALLBACK
    # called by TCP_pair start_notify{}, increase $actfl

    $self instvar actfl

    incr actfl
    
    # %%% DEBUG %%%
    # if { $stable } {  ; # load stable (GONG)
    #     global ns var flow_gen
    #     $self instvar pid
    #     $self instvar conns
    #     set fin_fid [$conns($connid) set fid]
    #     debugPrint $var(debug_tcl) 1 "--> $flow_gen/$var(sim_end) [format "%3.7f" [$ns now]] pid $pid cid $connid fid $fin_fid actfl $actfl pkts [format "%.f" $npkts] taskid $taskid"
    # }
    global ns var flow_gen
    $self instvar pid
    $self instvar conns
    set fin_fid [$conns($connid) set fid]
    debugPrint $var(debug_tcl) 1 "--> $flow_gen/$var(sim_end) [format "%3.7f" [$ns now]] pid $pid cid $connid fid $fin_fid actfl $actfl pkts [format "%.f" $npkts] taskid $taskid"
}

# p2-task (GONG)
# Agent_pair instproc fin_notify_ext { connid } {

#     $self instvar available
#     $self set available($connid) 1
# }

Agent_pair instproc fin_notify { connid dt taskid stable } { 
    # CALLBACK
    # called by TCP_pair fin_notify, decrease $actfl, increase $flow_fin, write $logfile, reschedule current connection if possible
    
    global ns var flow_gen flow_fin
    $self instvar pid
    $self instvar conns conns_per_pair
    $self instvar logfile
    $self instvar actfl
    
    if { $stable } {  ; # load stable (GONG)
        incr flow_fin
    }
    incr actfl -1 

    set fin_fid [$conns($connid) set fid]
    set fin_pkts [$conns($connid) set pkts]
    set fin_bps [expr $fin_pkts * $var(pkt_size_) * 8.0 / $dt]
    if { [info exists logfile] } { 
        if { $stable } { ; # load stable (GONG)
            puts $logfile "[format "%6.f" $fin_pkts] [format "%.7f" $dt] [format "%.4e" $fin_bps] $pid $connid $fin_fid"
            flush stdout
        }
    }
    
    # %%% DEBUG %%%
    #if { $stable } {  ; # load stable (GONG)
     #   debugPrint $var(debug_tcl) 1 "<-- $flow_fin/$var(sim_end) [format "%3.7f" [$ns now]] pid $pid cid $connid fid $fin_fid actfl $actfl pkts [format "%.f" $fin_pkts] dt [format "%.4e" $dt] bps [format "%.4e" $fin_bps] taskid $taskid"
    #}
	
	debugPrint $var(debug_tcl) 1 "<-- $flow_fin/$var(sim_end) [format "%3.7f" [$ns now]] pid $pid cid $connid fid $fin_fid actfl $actfl pkts [format "%.f" $fin_pkts] dt [format "%.4e" $dt] bps [format "%.4e" $fin_bps] taskid $taskid $stable"

    if { $flow_fin >= $var(sim_end) } {
        finish
    } 

    # p2-task (GONG) comment this out
    # if { $flow_gen < $var(sim_end) && $actfl < $conns_per_pair } {
    #     $self schedule $connid
    # }
    # p2-task (GONG)
    $self schedule $connid
}
