
#####################################################################
# MONITOR
#####################################################################

#:::::::::::::::: MONITOR: Queue Monitor ::::::::::::::::

proc debugPrint {mode newline msg} {
    # usage: debugPrint $var(debug_tcl) 1 "($i,$j,$init_fid_) "

    if { $mode == 1} {
        if { $newline == 1 } {
            puts $msg
        } else {
            puts -nonewline $msg
        }
        flush stdout
    }
}


array set qmon {}
array set qtrace {}
set fmon ""

proc setupQueueMonitor {mode output n1str n1 n2str n2} {
    # usage: setupQueueMonitor $var(debug_queue_mode) $output s$i $s($i) t$j $t($j)
    # output: 0 for internal use; 1 log to file

    global ns var
    upvar qmon queuemon
    upvar qtrace queuetrace
    upvar fmon flowmon

    # $mode == 0: monitor off
    if { $mode == 10 || $mode == 11 || $mode == 12 } {
        # queue monitor: ns-link.tcl, tools/queue-monitor.cc
        
        if { $output == 0 } {   ; # also for future manual trace
            set queuemon($n1str,$n2str) [$ns monitor-queue $n1 $n2 null $var(queue_sampling_interval_)]
        } else { 
            set queuemon($n1str,$n2str) [$ns monitor-queue $n1 $n2 [open queue_$n1str\_$n2str.tr w] $var(queue_sampling_interval_)]
        }

        if { $mode == 11 } {
            [$ns link $n1 $n2] start-tracing             ; # automatic trace all w/ event
        } elseif { $mode == 12 } {
            [$ns link $n1 $n2] queue-sample-timeout      ; # automatic trace all w/ interval
        } else {
            # $ns at 0.0 "queueTrace"                                 ; # schedule custom trace
            # [$ns link $n1 $n2] sample-queue-size                    ; # manual trace qavg
            # $qmon instvar parrivals_ pdepartures_ pdrops_ ...       ; # manual trace (bind)
            # $qmon get-bytes-integrator/get-pkts-integrator/get-delay-samples/printRTTs/printSeqnos/trace/traceDist                               ; # qmon commands
        }

    } elseif { $mode == 20 || $mode == 21 } {
        # queue trace: ns-lib.tcl
        set queuetrace($n1str,$n2str) [[$ns link $n1 $n2] queue]  ; # also for future manual trace
        $queuetrace($n1str,$n2str) attach [open queue_$n1str\_$n2str.tr w]

        if { $mode == 21 } {
            $queuetrace($n1str,$n2str) trace curq_      ; # automatic trace single bind value w/ event    
        } else {
            # $ns at 1 "queueTrace"                                 ; # schedule custom trace
            # set curq_ [$qtrace($n1str,$n2str) set curq_]          ; # manual trace (bind)
        }

    } elseif { $mode == 3 } {
        # flow monitor: ns-lib.tcl, tools/flowmon.cc
        set flowmon [$ns makeflowmon Fid]                   ; # SrcDstFid class coexistable
        $ns attach-fmon [$ns link $n1 $n2] $flowmon         ; # 3rd arg: 0/1 support early drop
        $flowmon attach [open queue_$n1str\_$n2str.tr w]
        # $ns at 1 "queueTrace"              ; # schedule custom trace
        # set fcl [$fmon classifier]         ; # can add SrcDstFid to choose
        # set flow [$fcl lookup auto 0 0 $i] ; # $i: fid_ of tcp
        # set bytes [$flow set bdepartures_] ; # manual trace (bind) check flow existence first!
        # $fmon dump/flows/classifier/attach ; # fmon commands
    }
}

#####################################################################
# LOAD CONVERGENCE CHECK
#####################################################################

#:::::::::::::::: load log ::::::::::::::::

# %%% DEBUG %%%
set loadlog [open load.tr w]
set recLoad {}
set isStable false      ; # stable indicator(help calculate real load during data collection)

proc queueTrace {} {
    global ns qmon num_servers_ var loadlog isStable; # %%% DEBUG %%% 
    global recLoad

    set lmload 0
    set lmthroughput 0
    
    for { set i 0 } { $i < $num_servers_ } { incr i } {
        set j [expr $i/$var(topo_spt)]

        incr lmload [$qmon(s$i,t$j) set parrivals_]
        $qmon(s$i,t$j) set parrivals_ 0

        incr lmthroughput [$qmon(t$j,s$i) set pdepartures_]
        $qmon(t$j,s$i) set pdepartures_ 0
    }

    # %%% DEBUG %%%
    puts $loadlog "[format "%.3f" [$ns now]] $lmload $lmthroughput $isStable"; flush stdout
    lappend recLoad $lmload

    $ns at [expr [$ns now] + $var(queue_sampling_interval_)] "queueTrace"
}

#:::::::::::::::: batch mean check ::::::::::::::::

# %%% DEBUG %%%
set statelog [open state.tr w]

set idxend 2            ; # end point of each check
set batchltot 2         ; # total batch length
set batchlen 0.02       ; # each batch length
set nbatch [expr int($batchltot/$batchlen)]     ; # number of batch 
set intval 0.1          ; # interval between each check
set stableth 0.07       ; # threshold of stable
# set isStable false      ; # stable indicator

proc stateCheck {} {
    global ns var idxend batchltot batchlen nbatch intval stableth isStable recLoad statelog ; # %%% DEBUG %%% 

    # set batchmeans {}
    # foreach idxint [range $nbatch] { 
    #     set totalrange [lrange $recLoad [expr int(($idxend - $batchltot + $idxint * $batchlen)*1000)] [expr int(($idxend - $batchltot + ($idxint + 1) * $batchlen)*1000-1)]]

    #     if { [llength $totalrange] > 0 } {
    #         set link_rate_pkt [expr $var(link_rate)*1e9/($var(pkt_size_) + 40)/8*$var(topo_spt)*$var(topo_tors)]
    #         set batchmean [expr [mean $totalrange] * 1000 / $link_rate_pkt ]
    #         lappend batchmeans $batchmean
    #     }
    # }
    # set sd [stddev $batchmeans]

    # # %%% DEBUG %%%
    # puts $statelog "[format "%.3f" [expr $idxend-$batchltot]] [format "%.3f" $idxend] [format "%.5f" $sd]"; flush stdout

    if { [$ns now] >= 4.0 } {
        set isStable true
        puts $statelog "[format "%.3f" [$ns now]] isStable=true"; flush stdout
    } else {
        set idxend [expr $idxend + $intval]
        $ns at [expr [$ns now] + $intval] "stateCheck"
        puts $statelog "[format "%.3f" [$ns now]] idxend: $idxend reschedule at [expr [$ns now] + $intval]"; flush stdout
    }

    # %%% AUTOMATIC %%%
    # if { $sd < $stableth } {
    #     set isStable true
    # } else {    ; # %%% DEBUG %%% # always schedule stateCheck to see curve
    #     set idxend [expr $idxend + $intval]
    #     $ns at [expr [$ns now] + $intval] "stateCheck"
    # }
}

#:::::::::::::::: helper ::::::::::::::::

proc lavg L {
    expr ([join $L +])/[llength $L].
}

proc mean list {
    set sum 0
    foreach i $list {set sum [expr {$sum+$i}]}
    expr {double($sum)/[llength $list]}
}

proc mean2 list {
    set sum 0
    foreach i $list {set sum [expr {$sum+$i*$i}]}
    expr {double($sum)/[llength $list]}
}

# set values {1 2 3 4 5}
# set SD [stddev $values]
proc stddev list {
    set m [lavg $list]
    expr {sqrt([mean2 $list]-$m*$m)}
} 

proc range args {
    foreach {start stop step} [switch -exact -- [llength $args] {
        1 {concat 0 $args 1}
        2 {concat   $args 1}
        3 {concat   $args  }
        default {error {wrong # of args: should be "range ?start? stop ?step?"}}
    }] break
    if {$step == 0} {error "cannot create a range when step == 0"}
    set range [list]
    while {$step > 0 ? $start < $stop : $stop < $start} {
        lappend range $start
        incr start $step
    }
    return $range
}