#!/usr/bin/perl 
no warnings;

#:::::::::::::::: helper ::::::::::::::::

$logfile = "sim.log"; 
open(LOG,">>$logfile") || die("Cannot Open Log File");
sub logprint { 
  my $str = shift; 
  print $str; 
  print LOG $str;
}

#:::::::::::::::: simulation config ::::::::::::::::

# assign profile dynamically or from profiles/
if ($#ARGV > -1) {
    @profiles = @ARGV;
} else {
    # @profiles = glob("profiles/*.tcl");           # with path
    @profiles = map m|([^/]+)$|, <profiles/*.tcl>;  # without path
}

# debug settings
$debug_ns = "/dev/null";     # toggle ns debug (error.tr, /dev/null) (to improve)
$trials = 1;               # repeat for each profile simulation


#:::::::::::::::: server & dir config :::::::::::::::: 

$user = "gong";

@server_workers = ("lame14","lame11","lame12","lame13","lame15","lame16");
$server_log = "lame14";     # special case in enst, all servers share the same storage

$dir_root = "/infres/nms/$user/ns3-repo/working";
$dir_ns = "/infres/nms/$user/ns3-repo/switch";
$dir_work = "$dir_root/working";
$dir_log = "$dir_root/log";

$num_cores = 30;        # number of cores to use each server (32 of each)
$wait_time = 10;        # interval when waiting availabe server/core


#####################################################################
# RUN
#####################################################################

foreach(@profiles){
    
    #:::::::::::::::: load profile & testbed :::::::::::::::: 

    # read first line of $profile.tcl to get testbed
    $profile = $_;
    open($fh, "profiles/$profile");
    $testbed = <$fh>;
    chomp $testbed;
    $testbed  = substr($testbed, 1);
    logprint "Running $profile on testbed: $testbed \n";

    #:::::::::::::::: ns enviornment :::::::::::::::: 

    $dir_tbd = "$dir_ns/$testbed";
    if ($testbed eq "codel") {
        $ns_path ="$dir_tbd/ns-allinone-2.36.rc1/bin:$dir_tbd/ns-allinone-2.36.rc1/tcl8.5.17/unix:$dir_tbd/ns-allinone-2.36.rc1/tk8.5.17/unix:\$PATH";
        $ld_lib_path = "$dir_tbd/ns-allinone-2.36.rc1/otcl-1.14:$dir_tbd/ns-allinone-2.36.rc1/lib";
        $tcl_lib = "$dir_tbd/ns-allinone-2.36.rc1/tcl8.5.17/library";
    } else {
        $ns_path ="$dir_tbd/ns-allinone-2.34/bin:$dir_tbd/ns-allinone-2.34/tcl8.4.18/unix:$dir_tbd/ns-allinone-2.34/tk8.4.18/unix:\$PATH";
        $ld_lib_path = "$dir_tbd/ns-allinone-2.34/otcl-1.13:$dir_tbd/ns-allinone-2.34/lib";
        $tcl_lib = "$dir_tbd/ns-allinone-2.34/tcl8.4.18/library";
    }
    $set_path_cmd = "export PATH=$ns_path && export LD_LIBRARY_PATH=$ld_lib_path && export TCL_LIBRARY=$tcl_lib";

    #:::::::::::::::: prepare log dir :::::::::::::::: 

    $dir_log_profile = "$dir_log/" . substr($profile, 0, -4);
    ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst)  = localtime(time);
    $mon += 1;
    $timestamp = sprintf("%02d%02d-%02d%02d%02d", $mon, $mday, $hour, $min, $sec);
    $dir_log_trials = "$dir_log_profile/$timestamp";

    `ssh $user\@$server_log 'mkdir $dir_log 2> /dev/null'`;
    `ssh $user\@$server_log 'mkdir $dir_log_profile 2> /dev/null'`;
    `ssh $user\@$server_log 'mkdir $dir_log_trials'`;

    #:::::::::::::::: repeating trials ::::::::::::::::

    for( my $trial = 0; $trial < $trials; $trial++ ){
    
        #:::::::::::::::: find available server/core ::::::::::::::::

        # indicate availability of server/core (current server_idx/core_idx)
        $available = 0; 
        $server_idx = 0;
        $core_idx = 0;

        while (!$available) {
            
            $server = $server_workers[$server_idx];        
            $dir_target = "$dir_work/$server-core$core_idx";
            
            # if dir_target exist -> test next
            # else available = true + mkdir dir_target
            `ssh $user\@$server 'test -e $dir_target'`;
            $rc = $? >> 8;

            #: $?>>8 the exit value of the subprocess (http://perldoc.perl.org/perlvar.html)  
            #: 'test -e file' return 0 if exist

            if ($rc) {
                $available = 1;
                `ssh $user\@$server 'mkdir $dir_target'`;
                last; # exits the loop
            }

            $core_idx++;
            if ($core_idx == $num_cores) {
                $core_idx = 0;                
                $server_idx++;
                if ($server_idx == @server_workers) {
                    $server_idx = 0;
                    logprint "Waiting for free server/cores for $wait_time sec..\n";
                    sleep($wait_time);
                }
            }
        }

        #:::::::::::::::: run sim on child process ::::::::::::::::

        logprint "Starting $timestamp-$trial on $server-$core_idx \n";

        `scp *.tcl $user\@$server:$dir_target/.`;
        `scp profiles/$profile $user\@$server:$dir_target/.`;

        $pid = fork();
        die "unable to fork: $!" unless defined($pid);
        if (!$pid) {  # child pid success
            
            `ssh $user\@$server '$set_path_cmd && cd $dir_target && nice ns sim.tcl profile $profile 2> $debug_ns 1> log.tr'`;

            #:::::::::::::::: post sim log record ::::::::::::::::

            logprint "Finishing $timestamp-$trial on $server-$core_idx => " . substr($profile, 0, -4) . "/$timestamp/$trial ... \n";

            `scp sim.config.tcl $user\@$server:$dir_log_trials/sim.config.tcl.$timestamp`;
            `ssh $user\@$server 'mv $dir_target/$profile $dir_log_trials/$profile.$timestamp'`;
            `ssh $user\@$server 'find $dir_target ! -name "*.tr" -type f -delete'`;
            `ssh $user\@$server 'mv $dir_target $dir_log_trials/$trial'`;

            exit(0);
        } else {
            sleep(1);
        }
    }
}

print "Waiting for child procs to terminate.\n";
while (wait() != -1) {
    # print "Waiting for child procs to terminate.\n";
}

$time = time - $^T;
logprint "Done ($time s)\n";
close(LOG);
