# add/remove packet headers as required
# this must be done before create simulator, i.e., [new Simulator]
# remove-all-packet-headers       ;# removes all except common
# add-packet-header Flags IP RCP  ;# PacketHeader/[hdr] at ~ns/tcl/lib/ns-packet.tcl

set ns [new Simulator]
#puts "Date: [clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S}]"
puts "Host: [exec hostname] [exec hostname -i]"
puts " "

#####################################################################
# ARGUMENTS
#####################################################################

set varname {profile}  

if { [expr [lsearch -exact $argv "help"] > -1 ]  } {
    puts [join $varname "\n"]
    exit 0
}

foreach name $varname {
    set index [lsearch -exact $argv $name]
    if { [expr $index > -1] } {
        set var($name) [lindex $argv [expr $index + 1]]
    }
}

source "sim.config.tcl"
source $var(profile)
source "sim.monitor.tcl"

#####################################################################
# NETWORK BUILDING
#####################################################################

source "common.tcl"

#:::::::::::::::: network ::::::::::::::::

puts "Setting up network ..."

set num_servers_ [expr $var(topo_spt) * $var(topo_tors)]
set uplink_cap_ [expr $var(link_rate) * $var(topo_spt) / $var(topo_spines) / $var(topo_x)]

# servers
for { set i 0 } { $i < $num_servers_ } { incr i } {
    set s($i) [$ns node]
}

# tors
for { set i 0 } { $i < $var(topo_tors) } { incr i } {
    set t($i) [$ns node]
}

# aggregators
for { set i 0 } { $i < $var(topo_spines) } { incr i } {
    set a($i) [$ns node]
}

# links servers <--> tors
for { set i 0 } { $i < $num_servers_ } { incr i } {
    set j [expr $i/$var(topo_spt)]
    $ns simplex-link $s($i) $t($j) [set var(link_rate)]Gb [expr $var(host_delay) + $var(link_delay)] $var(switch_alg)
    $ns simplex-link $t($j) $s($i) [set var(link_rate)]Gb [expr $var(host_delay) + $var(link_delay)] $var(switch_alg)

    # %%% DEBUG %%%
    setupQueueMonitor $var(debug_queue_mode) 0 s$i $s($i) t$j $t($j)
    setupQueueMonitor $var(debug_queue_mode) 0 t$j $t($j) s$i $s($i)
}

# links tors <--> aggregators
for { set i 0 } { $i < $var(topo_tors) } { incr i } {
    for { set j 0 } { $j < $var(topo_spines) } { incr j } {
        $ns duplex-link $t($i) $a($j) [set uplink_cap_]Gb $var(link_delay) $var(switch_alg)  
    }
}

puts "Network setup done.\n";flush stdout

#:::::::::::::::: Agents ::::::::::::::::

puts "Setting up agents ..."

set lambda [expr ($var(link_rate) * $var(load) * 1000000000) / ($var(mean_flow_size) * 8.0 / 1460 * 1500)]; # the expected number of "arrivals" that occur per unit time.
# flow arrival: poisson with inter-arrival [format "%.4f" [expr 1 / $lambda * 1000]] ms
puts "Lambda: $lambda\n";

set flowlog [open flow.tr w]

for { set i 0 } { $i < $num_servers_ } { incr i } {
    for { set j 0 } { $j < $num_servers_ } { incr j } {
        if { $j != $i } {

            set agents($i,$j) [new Agent_pair]

            $agents($i,$j) setup $s($i) $s($j) "$i $j" 1 "TCP_pair"
            $agents($i,$j) attach-logfile $flowlog
            
            $agents($i,$j) set_PCarrival_process  [expr $lambda / ($num_servers_ - 1)] "$var(workload).tcl" [expr int((1244 * $i + 17 * $j)*rand())] [expr int((4369 * $i + 33 * $j)*rand())]
        
            $ns at 0.0 "$agents($i,$j) init_schedule"
        }
    }
}
puts "Agents([expr $num_servers_ * ($num_servers_ - 1)]) setup done.\n"; flush stdout


#####################################################################
# RUN SIMULATION
#####################################################################

if { $var(debug_queue_mode) > 0 } {
    $ns at 0.0 "queueTrace" 
    $ns at $idxend "stateCheck"
} else {
    set isStable true
}

proc finish {} {
    global ns flowlog namfile

    $ns flush-trace

    if { [info exists flowlog] } { close $flowlog }

    if { [info exists loadlog] } { close $loadlog }
    if { [info exists statelog] } { close $statelog }

    if { [info exists namfile] } { 
        close $namfile
        exec nam out.nam &
    }

    puts "Simulation Finished!"

    exit 0
}

$ns at 1000.0 finish ; # time-based finish; extra large time to ensure finish

puts "Simulation started."

$ns run

